import logging
from .voices.transcribe.transcribe import Transcriber
from .voices.silences import SileroVoiceActivityDetection, VoiceActivityDetection
from .transcript.sentences import SentenceTokenizer

# Configure logger
LOG = logging.getLogger(__name__)

class Models:
    transcriber:Transcriber = None
    vad:VoiceActivityDetection = None
    sentence_tokenizer:SentenceTokenizer = None

    def __init__(
            self,
            device:str,
            whisper_model_size:str = "large-v3",
            stream:bool = False
        ):
        LOG.info("Initializing WhisperModel and audio stream.")
        # pylint: disable=import-outside-toplevel
        from .voices.transcribe.faster_whisper import FasterWhisperTranscriber
        self.transcriber:Transcriber = FasterWhisperTranscriber(model_size_or_path=whisper_model_size, device=device)

        LOG.info("Initializing Silero VAD model.")
        self.vad:VoiceActivityDetection = SileroVoiceActivityDetection()

        LOG.info("Initializing Sentence Tokenizer.")
        if stream:
            from .transcript.sentences import FastSentenceTokenizer
            self.sentence_tokenizer = FastSentenceTokenizer()
        else:
            from .transcript.sentences import SaTSentenceTokenizer
            self.sentence_tokenizer = SaTSentenceTokenizer(device)



import logging
import os
import sys
import wave
import traceback
from dataclasses import dataclass, field
from io import StringIO
from typing import List, Tuple, TextIO, Generator

import numpy as np
from colorama import Fore
from pyannote.core.annotation import Annotation

from .audio.sources.audiosource import AudioSource
from .transcript.words import VerbatimWord, VerbatimUtterance
from .audio.audio import samples_to_seconds
from .config import Config
from .models import Models
from .transcript.format.txt import (
    TranscriptFormatter,
    COLORSCHEME_ACKNOWLEDGED,
    COLORSCHEME_UNACKNOWLEDGED,
    COLORSCHEME_UNCONFIRMED
)
#pylint: disable=unused-import
from .voices.transcribe.transcribe import APPEND_PUNCTUATIONS, PREPEND_PUNCTUATIONS

# Configure logger
LOG = logging.getLogger(__name__)

@dataclass
class WhisperHistory:
    size: int = 12
    transcript_history: List[List[VerbatimWord]] = field(default_factory=list)

    def add(self, transcription: List[VerbatimWord]):
        self.transcript_history.append(transcription)
        if len(self.transcript_history) > self.size:
            self.transcript_history.pop(0)

    @staticmethod
    def advance_transcript(timestamp: int, transcript: List[VerbatimWord]) -> List[VerbatimWord]:
        return [w for w in transcript if w.end_ts > timestamp]

    def advance(self, timestamp: int):
        self.transcript_history = [self.advance_transcript(timestamp, transcript) for transcript in self.transcript_history]
        self.transcript_history = [h for h in self.transcript_history if len(h) > 0]

    @staticmethod
    def confirm_transcript(
        current_words:List[VerbatimWord],
        transcript:List[VerbatimWord],
        prefix:List[VerbatimWord],
        after_ts:int
    ) -> List[VerbatimWord]:
        c_index = 0
        while c_index < len(current_words):
            if (c_index < len(prefix) and prefix[c_index].word == current_words[c_index].word) or (current_words[c_index].start_ts <= after_ts):
                c_index = c_index + 1
            else:
                break

        p_index = 0
        while p_index < len(transcript):
            if (p_index < len(prefix) and prefix[p_index].word == transcript[p_index].word) or (transcript[p_index].start_ts <= after_ts):
                p_index = p_index + 1
            else:
                break

        confirmed_words = []

        while p_index < len(transcript) and c_index < len(current_words):
            if transcript[p_index].word.strip().lower() == current_words[c_index].word.strip().lower():
                LOG.debug(f"Confirming word '{transcript[p_index].word}'")
                confirmed_words += [current_words[c_index]]
                p_index += 1
                c_index += 1
            else:
                break
        if len(confirmed_words) > 0:
            LOG.debug(f"CONFIRMED CANDIDATE: {''.join([w.word for w in confirmed_words])}")
        return confirmed_words

    def confirm(self, current_words: List[VerbatimWord], prefix: List[VerbatimWord], after_ts: int) -> List[VerbatimWord]:
        return max((self.confirm_transcript(current_words=current_words, transcript=transcript, prefix=prefix, after_ts=after_ts)
                        for transcript in self.transcript_history),
                        key=len, default=[])

class RollingWindow:
    array:np.array
    def __init__(self, window_size, dtype=np.float32):
        self.reset(window_size=window_size, dtype=dtype)

    def reset(self, window_size=-1, dtype=None):
        if window_size == -1:
            window_size = len(self.array)
        if dtype is None:
            dtype = self.array.dtype
        self.array = np.zeros(window_size, dtype=dtype)

@dataclass
class State:
    confirmed_ts: int
    acknowledged_ts: int
    window_ts: int
    audio_ts: int
    transcript_candidate_history: WhisperHistory
    rolling_window:RollingWindow
    acknowledged_words:List[VerbatimWord]
    unconfirmed_words:List[VerbatimWord]
    acknowledged_utterances:List[VerbatimUtterance]
    unacknowledged_utterances:List[VerbatimUtterance]
    skip_silences: bool = True
    speaker_embeddings:List = None


    def __init__(self, config:Config):
        self.config = config
        self.confirmed_ts = -1
        self.acknowledged_ts = -1
        self.window_ts = 0  # Window timestamp in samples
        self.audio_ts = 0  # Global timestamp in samples
        self.transcript_candidate_history = WhisperHistory()
        self.acknowledged_words = []
        self.unconfirmed_words = []
        self.acknowledged_utterances = []
        self.unacknowledged_utterances = []
        self.speaker_embeddings = []

        window_size = self.config.sampling_rate * self.config.window_duration  # Total samples in 30 seconds
        self.rolling_window = RollingWindow(window_size=window_size, dtype=np.float32)  # Initialize empty rolling window

    def advance_audio_window(self, offset:int):
        if offset <= 0:
            return
        LOG.debug(f"Shifting rolling window by {offset} samples ({samples_to_seconds(offset)}s) "
                    "to offset {self.window_ts + offset} ({samples_to_seconds(self.window_ts + offset)}s).")

        self.rolling_window.array = np.roll(self.rolling_window.array, -offset)
        self.rolling_window.array[-offset:] = 0
        self.window_ts += offset

    def append_audio_to_window(self, audio_chunk:np.array):
        chunk_size = len(audio_chunk)
        window_size = len(self.rolling_window.array)
        if self.audio_ts + chunk_size <= self.window_ts + window_size:
            # Append to the current rolling window
            self.rolling_window.array[self.audio_ts - self.window_ts:self.audio_ts - self.window_ts + chunk_size] = audio_chunk
        else:
            # Shift the window and append new audio
            shift_amount = self.audio_ts + chunk_size - (self.window_ts + window_size)
            LOG.warning(f"failed to acknowledge transcription within audio buffer of {samples_to_seconds(window_size)} seconds")
            self.advance_audio_window(shift_amount)
            self.skip_silences = True
            self.rolling_window.array[-chunk_size:] = audio_chunk

        self.audio_ts += chunk_size


class Verbatim:
    state:State = None
    config:Config = None

    def __init__(self, config:Config, models=None):
        self.config = config
        self.state = State(config)
        if models is None:
            models = Models(device=config.device, stream=config.stream)
        self.models = models

        if config.start_time != 0:
            self.state.window_ts = config.start_time
            self.state.audio_ts = config.start_time

    def skip_leading_silence(self, min_speech_duration_ms:int = 500) -> int:
        min_speech_duration_ms = 750
        min_speech_duration_samples = 16000 * min_speech_duration_ms // 1000
        audio_samples = self.state.audio_ts - self.state.window_ts
        voice_segments = self.models.vad.find_activity(
            audio=self.state.rolling_window.array[0:audio_samples],
            min_speech_duration_ms=min_speech_duration_ms)
        LOG.debug(f"Voice segments: {voice_segments}")
        if len(voice_segments) == 0:
            LOG.info(f"Skipping silences between {samples_to_seconds(self.state.window_ts):.2f} and {samples_to_seconds(self.state.audio_ts):.2f}")

            # preserve a bit of audio at the end that may have been too short just because it is truncated
            padding = min_speech_duration_samples
            remaining_audio = self.state.audio_ts - self.state.window_ts - padding
            self.state.advance_audio_window(remaining_audio)

            return self.state.audio_ts

        voice_start = voice_segments[0]['start']
        voice_end = voice_segments[0]['end']
        voice_length = voice_end - voice_start

        # skip leading silences
        if voice_start > 0:
            LOG.info(f"Skipping silences between {samples_to_seconds(self.state.window_ts):.2f}"
                     f" and {samples_to_seconds(self.state.window_ts + voice_start):.2f}")
            self.state.advance_audio_window(voice_start)
            return voice_length + self.state.window_ts

        return voice_length + self.state.window_ts

    def dump_window_to_file(self, filename: str = "debug_window.wav"):
        window = self.state.rolling_window.array
        LOG.debug("Dumping current window to file for debugging.")
        # pylint: disable=no-member
        with wave.open(filename, 'wb') as wf:
            wf.setnchannels(1)  # Mono
            wf.setsampwidth(2)  # 16-bit audio
            wf.setframerate(self.config.sampling_rate)
            wf.writeframes((window * 32768.0).astype(np.int16).tobytes())
        LOG.debug("Finished writing window to file.")


    @staticmethod
    def align_words_to_sentences(sentences: list[str], window_words: list[VerbatimWord]) -> list[VerbatimUtterance]:
        # Concatenate all sentences into a single string
        full_text_sentences = ''.join(sentences)
        # Concatenate all words from window_words
        full_text_words = ''.join(w.word for w in window_words)

        # Basic validation step (optional but recommended)
        if full_text_sentences != full_text_words:
            raise ValueError("The joined text from sentences and window_words do not match.")

        # We'll iterate through sentences and pick words until we match them character-for-character.
        result = []
        current_char_index = 0
        current_word_index = 0

        def remove_spaces_and_punctuation(string: str) -> str:
            return string.translate(str.maketrans('', '', " " + PREPEND_PUNCTUATIONS + APPEND_PUNCTUATIONS))


        # For each sentence, we want to collect a sublist of VerbatimWord
        for sentence in sentences:
            # igonre punctuation and spaces
            sentence = remove_spaces_and_punctuation(sentence)
            sentence_length = len(sentence.strip())
            target_end = current_char_index + sentence_length

            sentence_words = []
            # Accumulate words until we have accounted for all chars of this sentence
            while current_char_index < target_end:
                # Pick the next word
                w = window_words[current_word_index]
                sentence_words.append(w)
                word_text = remove_spaces_and_punctuation(w.word)
                current_char_index += len(word_text)
                current_word_index += 1

                # If we overshoot, something is wrong,
                # but given the perfect alignment we expect to land exactly on target_end
                if current_char_index > target_end:
                    raise ValueError("Mismatch in alignment between sentences and words.")

            # Now we have all the words for this sentence
            if len(sentence_words) > 0:
                result.append(VerbatimUtterance.from_words(sentence_words))

        # At the end, `result` should be a List[List[VerbatimWord]]
        return result


    @staticmethod
    def words_to_sentences(word_tokenizer, window_words:List[VerbatimWord]) -> list[VerbatimUtterance]:
        sentences = []
        if len(window_words) == 0:
            return []

        window_text = ''.join([w.word for w in window_words])
        for tok in word_tokenizer.split(window_text):
            sentences += [tok]

        utterances = Verbatim.align_words_to_sentences(sentences=sentences, window_words=window_words)
        return utterances

    def _guess_language(self, audio:np.array, sample_offset:int, sample_duration:int, lang:List[str]) -> Tuple[str,float]:
        lang_samples = audio[sample_offset:sample_offset + sample_duration]
        LOG.info(f"Detecting language using samples {sample_offset}({samples_to_seconds(sample_offset)}) "
                 f"to {sample_offset + sample_duration}({samples_to_seconds(sample_offset + sample_duration)})")
        return self.models.transcriber.guess_language(audio=lang_samples, lang=lang)


    def guess_language(self, timestamp:int):

        if len(self.config.lang) == 0:
            LOG.warning("Language is not set - defaulting to english")
            return "en"

        if len(self.config.lang) == 1:
            return self.config.lang[0]

        lang_sample_start = max(0, timestamp - self.state.window_ts)
        available_samples = self.state.audio_ts - self.state.window_ts - lang_sample_start
        lang_samples_size = min(2 * 16000, available_samples)

        while True:
            lang, prob = self._guess_language(
                audio=self.state.rolling_window.array,
                sample_offset=lang_sample_start,
                sample_duration=lang_samples_size,
                lang = self.config.lang
                )
            if prob > 0.5 or lang_samples_size == available_samples:
                break
            # retry with larger sample
            lang_samples_size = min(2 * lang_samples_size, available_samples)

        return lang

    def transcribe_window(self) -> Tuple[List[VerbatimWord], List[VerbatimWord]]:
        LOG.info("Starting transcription of audio chunk.")
        LOG.info(f"Window Start Time: {samples_to_seconds(self.state.window_ts)}")
        LOG.info(f"Confirmed Time: {samples_to_seconds(self.state.confirmed_ts)}")
        LOG.info(f"Acknowledged Time: {samples_to_seconds(self.state.acknowledged_ts)}")
        LOG.info(f"Valid audio range: 0.0 - {samples_to_seconds(self.state.audio_ts - self.state.window_ts)}")

        acknowledged_words_in_window = WhisperHistory.advance_transcript(timestamp=self.state.window_ts, transcript=self.state.acknowledged_words)
        lang = self.guess_language(timestamp=max(0, self.state.acknowledged_ts))
        prefix_text = ""
        for word in [w for u in self.state.unacknowledged_utterances for w in u.words]:
            if word.lang != lang:
                break
            prefix_text += word.word
        whisper_prompt = self.config.whisper_prompts[lang] if lang in self.config.whisper_prompts else self.config.whisper_prompts["en"]
        transcript_words = self.models.transcriber.transcribe(
            audio=self.state.rolling_window.array,
            lang=lang, prompt=whisper_prompt, prefix=prefix_text,
            window_ts=self.state.window_ts, audio_ts=self.state.audio_ts,
            whisper_beam_size = self.config.whisper_beam_size,
            whisper_best_of = self.config.whisper_best_of,
            whisper_patience = self.config.whisper_patience,
            whisper_temperatures = self.config.whisper_temperatures,
            )

        self.state.transcript_candidate_history.advance(self.state.window_ts)
        confirmed_words = self.state.transcript_candidate_history.confirm(
            current_words=transcript_words, after_ts=self.state.acknowledged_ts-1, prefix=acknowledged_words_in_window)
        self.state.transcript_candidate_history.add(transcription=transcript_words)

        if len(confirmed_words) > 0:
            self.state.confirmed_ts = confirmed_words[-1].start_ts
            LOG.debug(f"Confirmed ts: {self.state.confirmed_ts} ({samples_to_seconds(self.state.confirmed_ts)})")

            newline = os.linesep
            transcript_history = self.state.transcript_candidate_history.transcript_history
            LOG.debug(f"Transcript:\n{newline.join([''.join([w.word for w in history]) for history in transcript_history])}\n"
                        f"{''.join(w.word for w in transcript_words)}\n{''.join(w.word for w in confirmed_words)}")

        unconfirmed_words = [transcript_words[i] for i in range(len(confirmed_words), len(transcript_words))]
        return confirmed_words, unconfirmed_words

    def get_next_number_of_chunks(self):
        available_chunks = self.config.window_duration - float(self.state.audio_ts - self.state.window_ts) / self.config.sampling_rate

        thresholds = self.config.chunk_table

        for limit, value in thresholds:
            limit_sample = limit * self.config.window_duration
            value_sample = value * self.config.window_duration
            if available_chunks >= limit_sample:
                return value_sample

        # If for some reason available_chunks is less than 0, return the smallest chunk count.
        return 1

    def pretty_print_transcript(
        self,
        acknowledged_utterances:List[VerbatimUtterance],
        unacknowledged_utterances:List[VerbatimUtterance],
        unconfirmed_words:List[VerbatimWord],
        file:TextIO = sys.stdout
    ):
        formatter:TranscriptFormatter = TranscriptFormatter()
        file.write(
            f"[{samples_to_seconds(self.state.window_ts)}/"
            f"{samples_to_seconds(self.state.audio_ts - self.state.acknowledged_ts)}/"
            f"{samples_to_seconds(self.state.audio_ts - self.state.confirmed_ts)}]" + Fore.LIGHTGREEN_EX)
        for u in acknowledged_utterances:
            formatter.format_utterance(utterance=u, out=file, colours=COLORSCHEME_ACKNOWLEDGED)
        for u in unacknowledged_utterances:
            formatter.format_utterance(utterance=u, out=file, colours=COLORSCHEME_UNACKNOWLEDGED)
        if len(unconfirmed_words) > 0:
            formatter.format_utterance(utterance=VerbatimUtterance.from_words(unconfirmed_words), out=file, colours=COLORSCHEME_UNCONFIRMED)
        file.write(os.linesep)
        file.flush()

    def acknowledge_utterances(
            self,
            utterances:List[VerbatimUtterance],
            min_ack_duration=16000, min_unack_duration=16000,
    ) -> Tuple[List[VerbatimUtterance], List[VerbatimUtterance]]:
        if len(utterances) == 0:
            return [], utterances

        # check if the last utterance is complete
        last_valid_utterance_index = len(utterances) - 1
        valid_endings = tuple(APPEND_PUNCTUATIONS)
        if not utterances[last_valid_utterance_index].text.endswith(valid_endings):
            last_valid_utterance_index = last_valid_utterance_index - 1

        if last_valid_utterance_index < 0:
            return [], utterances

        start_ts = utterances[0].start_ts
        max_duration = self.state.audio_ts - self.state.window_ts - min_unack_duration

        duration = 0
        index = 0
        while duration < min_ack_duration and index <= last_valid_utterance_index:
            new_duration = utterances[index].end_ts - start_ts
            if new_duration > max_duration:
                break
            duration = new_duration
            index += 1

        return utterances[0:index], utterances[index:]

    def get_speaker_at(self, time:float, diarization:Annotation):
        if diarization is None:
            return None

        for turn, _, speaker in diarization.itertracks(yield_label=True):
            if turn.end < time:
                continue
            if turn.start > time:
                break
            return speaker
        return None

    def get_speaker_before(self, time:float, diarization:Annotation):
        last_speaker = None
        for turn, _, speaker in diarization.itertracks(yield_label=True):
            if turn.end < time:
                last_speaker = speaker
                continue
            if turn.start > time:
                break
        return last_speaker

    def get_speaker_after(self, time:float, diarization:Annotation):
        for turn, _, speaker in diarization.itertracks(yield_label=True):
            if turn.start > time:
                return speaker
        return None

    def assign_speaker(self, utterance:VerbatimUtterance, diarization:Annotation):
        if diarization is None:
            return None

        start = utterance.start_ts / 16000.0
        end =utterance.end_ts / 16000.0
        duration = end - start
        samples:List[float]
        if duration < 1:
            samples = [ start + duration * 0.5]
        elif duration < 4:
            samples = [ start + duration * 0.25,
                        start + duration * 0.5,
                        start + duration * 0.75 ]
        else:
            samples = [ start + duration * 0.2,
                        start + duration * 0.3,
                        start + duration * 0.4,
                        start + duration * 0.5,
                        start + duration * 0.6,
                        start + duration * 0.7,
                        start + duration * 0.8 ]

        votes = {}
        for sample in samples:
            speaker = self.get_speaker_at(time=sample, diarization=diarization)
            if speaker:
                if speaker in votes:
                    votes[speaker] = votes[speaker] + 1
                else:
                    votes[speaker] = 1

        if len(votes) != 0:
            best_votes = 0
            best_speaker = ""
            for speaker, vote in votes.items():
                if vote > best_votes:
                    best_speaker = speaker
                    best_votes = vote
            return best_speaker

        speaker = self.get_speaker_before(time=start + duration * 0.5, diarization=diarization)
        if speaker:
            return speaker

        speaker = self.get_speaker_after(time=start + duration * 0.5, diarization=diarization)
        if speaker:
            return speaker

        return None

    def process_audio_window(self) -> Generator[Tuple[VerbatimUtterance,List[VerbatimUtterance],List[VerbatimWord]], None, None]:
        while True:
            # minimum number of samples to attempt transcription
            min_audio_duration_samples = 16000
            min_speech_duration_ms = 500
            utterances = []
            enable_vad = True
            if enable_vad and self.state.skip_silences:
                self.state.skip_silences = False
                self.skip_leading_silence(min_speech_duration_ms=min_speech_duration_ms)
                if self.state.audio_ts - self.state.window_ts < min_audio_duration_samples:
                    # we skipped all available audio - keep skipping silences and do nothing else for now
                    self.state.skip_silences = True

            if self.state.audio_ts - self.state.window_ts < min_audio_duration_samples:
                return

            if self.config.debug:
                self.dump_window_to_file(filename=f"{self.config.working_prefix_no_ext}-debug_window.wav")  # Dump current window for debugging

            confirmed_words, unconfirmed_words = self.transcribe_window()
            self.state.unconfirmed_words = unconfirmed_words
            if len(confirmed_words) > 0:
                utterances = self.words_to_sentences(word_tokenizer=self.models.sentence_tokenizer, window_words=confirmed_words)
                acknowledged_utterances, confirmed_utterances = self.acknowledge_utterances(utterances=utterances)

                if self.config.diarization:
                    for acknowledged_utterance in acknowledged_utterances:
                        acknowledged_utterance.speaker = self.assign_speaker(acknowledged_utterance, self.config.diarization)
                #elif self.config.stream and self.config.diarize:
                #    for acknowledged_utterance in acknowledged_utterances:
                #        self.diarize_utterance(acknowledged_utterance)

                self.state.acknowledged_utterances += acknowledged_utterances
                self.state.unacknowledged_utterances = confirmed_utterances

                for i, utterance in enumerate(acknowledged_utterances):
                    yield utterance, acknowledged_utterances[i+1:] + confirmed_utterances, unconfirmed_words

                if len(acknowledged_utterances) > 0:
                    for u in acknowledged_utterances:
                        self.state.acknowledged_words += u.words

                    # utterances are split at short pauses, advance a bit to avoid repeating the last word
                    # but not too much as to skip the first word of the next utterance
                    utterance_padding_ms = 100
                    utterance_padding_samples = utterance_padding_ms * 16000 // 1000

                    self.state.acknowledged_ts = acknowledged_utterances[-1].end_ts + utterance_padding_samples
                    self.state.skip_silences = True
            else:
                acknowledged_utterances = []
                confirmed_utterances = []


            outstr = StringIO()
            self.pretty_print_transcript(
                acknowledged_utterances=[], unacknowledged_utterances=confirmed_utterances, unconfirmed_words=unconfirmed_words, file=outstr)
            LOG.info(outstr.getvalue())

            self.state.acknowledged_words = WhisperHistory.advance_transcript(
                timestamp=self.state.window_ts, transcript=self.state.acknowledged_words)

            if self.state.acknowledged_ts > self.state.window_ts:
                shift_amount = self.state.acknowledged_ts - self.state.window_ts
                self.state.advance_audio_window(shift_amount)
                self.state.skip_silences = True

            if len(acknowledged_utterances) <= 1:
                break

    def capture_audio(self, audio_source:AudioSource):
        if not audio_source.has_more():
            return False

        next_chunk = self.get_next_number_of_chunks()
        LOG.debug(f"capturing {next_chunk} chunks of audio")
        audio_array = audio_source.next_chunk(next_chunk)
        self.state.append_audio_to_window(audio_array)
        return True

    def transcribe(self) -> Generator[Tuple[VerbatimUtterance,List[VerbatimUtterance],List[VerbatimWord]], None, None]:
        self.state.rolling_window.reset()  # Initialize empty rolling window

        try:
            self.config.source_stream.open()
            LOG.info("Starting main loop for audio transcription.")
            while True:
                has_more_audio = self.capture_audio(audio_source=self.config.source_stream)
                had_utterances = False

                # capture any utterance that slipped out of the current window
                flushed_utterances = []
                while len(self.state.unacknowledged_utterances) > 0:
                    if self.state.unacknowledged_utterances[0].end_ts > self.state.window_ts:
                        break
                    utterance = self.state.unacknowledged_utterances.pop(0)
                    utterance.speaker = self.assign_speaker(utterance, self.config.diarization)
                    yield utterance, self.state.unacknowledged_utterances, self.state.unconfirmed_words

                if len(self.state.unacknowledged_utterances) > 0:
                    flushed_utterances_words = []
                    partial_utterance = self.state.unacknowledged_utterances[0]
                    while len(partial_utterance.words) > 0:
                        if partial_utterance.words[0].end_ts > self.state.window_ts:
                            break
                        flushed_word = partial_utterance.words.pop(0)
                        flushed_utterances_words.append(flushed_word)
                        partial_utterance.start_ts = partial_utterance.words[-1].start_ts
                        partial_utterance.text = [w.word for w in partial_utterance.words]

                    if len(flushed_utterances_words) > 0:
                        utterance = VerbatimUtterance.from_words(flushed_utterances_words)
                        utterance.speaker = self.assign_speaker(utterance, self.config.diarization)
                        yield utterance, self.state.unacknowledged_utterances, self.state.unconfirmed_words
                else:
                    flushed_utterances_words = []
                    while len(self.state.unconfirmed_words) > 0:
                        if self.state.unconfirmed_words[0].end_ts > self.state.window_ts:
                            break
                        flushed_word = self.state.unconfirmed_words.pop(0)
                        flushed_utterances_words.append(flushed_word)

                    if len(flushed_utterances_words) > 0:
                        utterance = VerbatimUtterance.from_words(flushed_utterances_words)
                        utterance.speaker = self.assign_speaker(utterance, self.config.diarization)
                        yield utterance, self.state.unacknowledged_utterances, self.state.unconfirmed_words

                if len(flushed_utterances_words) > 0:
                    flushed_utterances.append(VerbatimUtterance.from_words(flushed_utterances_words))

                for utterance, unacknowmedged, unconfirmed in self.process_audio_window():
                    had_utterances = True
                    yield utterance, unacknowmedged, unconfirmed

                if not had_utterances and not has_more_audio:
                    break

        except KeyboardInterrupt:
            LOG.info("KeyboardInterrupt detected, stopping transcription.")
            LOG.debug("Stopping...")
        # pylint: disable=broad-exception-caught
        except Exception as e:
            LOG.error(f"An unexpected error occurred: {e}\n{traceback.format_exc()}")
            LOG.debug("Stopping...")
        finally:
            LOG.info("Cleaning up resources.")

            for i, utterance in enumerate(self.state.unacknowledged_utterances):
                utterance.speaker = self.assign_speaker(utterance, self.config.diarization)
                yield utterance, self.state.unacknowledged_utterances[i+1:], self.state.unconfirmed_words

            if len(self.state.unconfirmed_words) > 0:
                unconfirmed_utterance:VerbatimUtterance = VerbatimUtterance.from_words(self.state.unconfirmed_words)
                unconfirmed_utterance.speaker = self.assign_speaker(unconfirmed_utterance, self.config.diarization)
                yield unconfirmed_utterance, [], []

            self.config.source_stream.close()



from dataclasses import dataclass, field
from typing import List

from faster_whisper.transcribe import Word

from ..audio.audio import samples_to_seconds


@dataclass
class VerbatimWord:
    start_ts: int
    end_ts: int
    word: str
    probability: float
    lang: str

    @classmethod
    def from_word(cls, word: Word, lang:str, ts_offset: int = 0) -> "VerbatimWord":
        """Creates a VerbatimWord instance from a Word object with a timestamp offset."""
        start_ts = int(word.start * 16000) + ts_offset
        end_ts = int(word.end * 16000) + ts_offset
        return cls(start_ts=start_ts, lang=lang, end_ts=end_ts, word=word.word, probability=word.probability)

@dataclass
class VerbatimUtterance:
    speaker:str
    start_ts: int
    end_ts: int
    text:str
    words:List[VerbatimWord] = field(default_factory=list)

    def get_start(self) -> float:
        return samples_to_seconds(self.start_ts)

    def get_end(self) -> float:
        return samples_to_seconds(self.end_ts)

    @classmethod
    def from_words(cls, words:List[VerbatimWord], speaker:str=None) -> "VerbatimUtterance":
        start_ts = words[0].start_ts
        end_ts = words[-1].end_ts
        text = ''.join([w.word for w in words])
        return cls(start_ts=start_ts, end_ts=end_ts, words=words, text=text, speaker=speaker)



import logging

import numpy as np
import scipy.io.wavfile
import torch
from pyannote.audio import Inference, Model, Pipeline
from pyannote.audio.pipelines.utils.hook import ProgressHook
from pyannote.audio.utils.multi_task import map_with_specifications
from pyannote.audio.utils.reproducibility import fix_reproducibility
from pyannote.core.annotation import Annotation
from pyannote.database.util import load_rttm
from scipy.spatial.distance import cdist

from ..audio.audio import wav_to_int16
from ..transcript.words import VerbatimUtterance

# Configure logger
LOG = logging.getLogger(__name__)

class Diarization:
    def __init__(self, device:str, huggingface_token:str, use_ami:bool= False):
        LOG.info("Initializing Diarization Pipeline.")
        self.huggingface_token = huggingface_token
        self._use_ami = use_ami
        if self._use_ami:
            self.pipeline = Pipeline.from_pretrained(
                "pyannote/speech-separation-ami-1.0",
                use_auth_token=self.huggingface_token
            )
            hyper_parameters =         {
                    "segmentation": {
                    "min_duration_off": 0.0,
                    "threshold": 0.82
                    },
                    "clustering": {
                        "method": "centroid",
                        "min_cluster_size": 15,
                        "threshold": 0.68,
                    },
                    "separation": {
                        "leakage_removal": True,
                        "asr_collar": 0.32,
                    }
                }

            self.pipeline.instantiate(hyper_parameters)
        else:
            self.pipeline = Pipeline.from_pretrained(
                checkpoint_path="pyannote/speaker-diarization-3.1",
                use_auth_token=self.huggingface_token
            )
            self.pipeline.instantiate({})

        self.pipeline.to(torch.device(device))

    @staticmethod
    def load_diarization(rttm_file:str):
        rttms = load_rttm(file_rttm=rttm_file)
        annotation:Annotation = next(iter(rttms.values()))
        return annotation

    # pylint: disable=unused-argument
    def compute_diarization(self, file_path:str, out_rttm_file:str = None, nb_speakers:int=None) -> Annotation:
        if not out_rttm_file:
            out_rttm_file = "out.rttm"

        sources = None
        with ProgressHook() as hook:
            if self._use_ami:
                diarization, sources = self.pipeline(file_path, hook=hook)
            else:
                diarization = self.pipeline(file_path, hook=hook, num_speakers=nb_speakers)

        # dump the diarization output to disk using RTTM format
        with open(out_rttm_file, "w", encoding="utf-8") as rttm:
            diarization.write_rttm(rttm)

        if sources:
            # dump sources to disk as SPEAKER_XX.wav files
            for s, speaker in enumerate(diarization.labels()):
                if s < sources.data.shape[1]:
                    speaker_data = sources.data[:, s]
                    if speaker_data.dtype != np.int16:
                        speaker_data = wav_to_int16(speaker_data)
                    scipy.io.wavfile.write(f'{speaker}.wav', 16000, speaker_data)
                else:
                    LOG.debug(f"Skipping speaker {s} as it is out of bounds.")
        return diarization

    def diarize_utterance(self, u:VerbatimUtterance, window_ts:int, audio:np.array, speaker_embeddings):
        rel_start = max(0, u.start_ts - window_ts)
        rel_end = max(0, u.end_ts - window_ts)
        # compute embedding and compare
        model = Model.from_pretrained(checkpoint="pyannote/embedding",
                                      use_auth_token=self.huggingface_token)

        inference = Inference(model, window="whole")
        fix_reproducibility(inference.device)
        inference.to(torch.device("cuda"))
        audio_segment = torch.tensor(audio[rel_start:rel_end].reshape(1, 1, -1),
                                     device='cuda')

        output = inference.infer(audio_segment)

        # pylint: disable=unused-argument
        def __first_sample(outputs: np.ndarray, **kwargs) -> np.ndarray:
            return outputs[0]

        embedding = map_with_specifications(inference.model.specifications, __first_sample, output)
        embedding = embedding.reshape(1, -1)

        best_i = -1
        best_distance = 1
        for i, speaker_embedding in enumerate(speaker_embeddings):
            distance = cdist(embedding, speaker_embedding, metric="cosine")[0, 0]
            if distance < best_distance:
                best_distance = distance
                best_i = i
        if best_i == -1:
            speaker_embeddings.append(embedding)
            print("Added first speaker")
        else:
            if best_distance < 0.15:
                print(f"Detected speaker {best_i}")
            else:
                print(f"Added speaker {best_i}")
                speaker_embeddings.append(embedding)



import logging
from typing import Tuple

import librosa
import numpy as np
from audio_separator.separator import Separator
from scipy.io.wavfile import write as wav_write

from ..audio.audio import format_audio

# Configure logger
LOG = logging.getLogger(__name__)

class VoiceSeparator:
    def __init__(self, log_level:int = logging.WARN, model_name:str = None):
        if model_name is None:
            #model_name = 'MDX23C-8KFFT-InstVoc_HQ.ckpt'
            model_name = 'MDX23C-8KFFT-InstVoc_HQ_2.ckpt'
        self.separator = Separator(log_level=log_level, sample_rate=16000)
        self.separator.load_model(model_name)

    def isolate_voice_in_file(self, file:str, out_voice:str = None, out_noise:str = None) -> Tuple[str, str]:
        if out_voice is None:
            out_voice = "debug_audio-vocals"
        if out_noise is None:
            out_noise = "debug_audio-instrumental"

        # Use MDX to separate vocals from the source audio
        output_file_paths = self.separator.model_instance.separate(file, {
            "Instrumental": out_noise,
            "Vocals": out_voice,
        })

        # The separated files: instrument track and vocal track
        instrument_audio_path = output_file_paths[0]
        voice_audio_path = output_file_paths[1]

        return voice_audio_path, instrument_audio_path

    def isolate_voice_in_array(self, audio: np.array) -> np.array:
        input_length = len(audio)

        if not np.any(audio):
            # audio is empty, skip
            return audio

        # Save the input audio to a temporary file
        temp_audio_file = "voice-isolation.wav"
        wav_write(temp_audio_file, 16000, audio)

        voice_audio_path, _ = self.isolate_voice_in_file(file=temp_audio_file)

        # Load the vocal audio back into a NumPy array
        voice_audio, voice_sampling_rate = librosa.load(voice_audio_path, sr=None, mono=False)
        voice_audio = voice_audio.T # librosa formats (nchannel, samples) and we expect (samples, nchannels)

        # Format the vocal audio to mono and 16 kHz
        formatted_voice_audio = format_audio(voice_audio, voice_sampling_rate)
        wav_write("voice-isolation-filtered.wav", 16000, formatted_voice_audio)

        # ensure output has same length as input
        if len(formatted_voice_audio) < input_length:
            padding_length = input_length - formatted_voice_audio
            formatted_voice_audio = np.pad(formatted_voice_audio, (0, padding_length), mode='constant', constant_values=0)
        elif len(formatted_voice_audio) > input_length:
            formatted_voice_audio = formatted_voice_audio[:input_length]

        return formatted_voice_audio



import logging
from typing import List, Tuple, Union

import numpy as np
import whisper
from whisper.model import Whisper

from .transcribe import Transcriber, WhisperConfig
from ...transcript.words import VerbatimWord

LOG = logging.getLogger(__name__)


class WhisperTranscriber(Transcriber):
    def __init__(self, *, model_size_or_path:str, device:str,
        whisper_beam_size:int = 3,
        whisper_best_of:int = 3,
        whisper_patience:float = 1.0,
        whisper_temperatures:List[float] = None
    ):
        if whisper_temperatures is None:
            whisper_temperatures = [0.0, 0.2, 0.4, 0.6, 0.8, 1.0]

        self.device = device
        self.whisper_beam_size = whisper_beam_size
        self.whisper_best_of = whisper_best_of
        self.whisper_patience = whisper_patience
        self.whisper_temperatures = whisper_temperatures
        self.model:Whisper = whisper.load_model(model_size_or_path, device=device)


    def guess_language(self, audio:np.array, lang:List[str]) -> Tuple[str, float]:
        audio = whisper.pad_or_trim(audio)
        mel_spectrogram = whisper.log_mel_spectrogram(audio, n_mels=self.model.dims.n_mels).to(self.model.device)

        _, lang_probs = self.model.detect_language(mel=mel_spectrogram)
        best_lang = max((k for k in lang_probs if k in lang), key=lang_probs.get)

        return best_lang, lang_probs[best_lang]

    def transcribe(
            self, *,
            audio:np.array,
            lang: str, prompt: str, prefix: str,
            window_ts:int, audio_ts:int,
            whisper_beam_size: int = 3,
            whisper_best_of: int = 3,
            whisper_patience: float = 1.0,
            whisper_temperatures: List[float] = None,
    ) -> List[VerbatimWord]:
        if whisper_temperatures is None:
            whisper_temperatures = [0.0, 0.2, 0.4, 0.6, 0.8, 1.0]

        use_fp16 = self.device == "cuda"
        verbose:Union[None,bool] = None
        if LOG.getEffectiveLevel() < logging.INFO:
            verbose = True
        elif LOG.getEffectiveLevel() < logging.WARN:
            verbose = False # will still display a progress bar
        else:
            verbose = None

        whisper_config:WhisperConfig = WhisperConfig()

        options = whisper.DecodingOptions(
            task=whisper_config.task,
            language=lang,
            temperature=tuple(whisper_temperatures),
            sample_len=None,
            best_of=whisper_best_of,
            beam_size=whisper_beam_size,
            patience=whisper_patience,
            length_penalty=whisper_config.length_penalty,
            prompt=prompt,
            prefix=prefix,
            suppress_tokens=whisper_config.suppress_tokens,
            suppress_blank=whisper_config.suppress_blank,
            without_timestamps=False,
            max_initial_timestamp=1.0,
            fp16=use_fp16
        )

        transcript = self.model.transcribe(
            audio=audio,
            word_timestamps=True,
            verbose=verbose,
            compression_ratio_threshold= whisper_config.compression_ratio_threshold,
            logprob_threshold = whisper_config.logprob_threshold,
            no_speech_threshold = whisper_config.no_speech_threshold,
            condition_on_previous_text=False,
            initial_prompt=prompt,
            prepend_punctuations=whisper_config.prepend_punctuations,
            append_punctuations=whisper_config.append_punctuations,
            clip_timestamps=[0.0],
            hallucination_silence_threshold=None,
            **options.__dict__,
        )
        words:List[VerbatimWord] = []
        for segment in transcript['segments']:
            # pylint: disable=unused-variable
            segment_id:int = segment.get('id')
            segment_seek:int = segment.get('seek')
            segment_start:str = segment.get('start')
            segment_end:str = segment.get('end')
            segment_text:str = segment.get('text')
            segment_temperature:float = segment.get('temperature')
            segment_avg_logprob:float = segment.get('avg_logprob')
            segment_compression_ratio:float = segment.get('compression_ratio')
            segment_no_speech_prob:float = segment.get('no_speech_prob')
            segment_words:List[str,object] = segment.get('words')
            for word in segment_words:
                word_start:float = word.get('start')
                word_end:float = word.get('end')
                word_text:str = word.get('word')
                word_probability:float = word.get('probability')

                start_ts = int(word_start*16000) + window_ts
                end_ts = int(word_end*16000) + window_ts
                words.append(VerbatimWord(start_ts=start_ts, end_ts=end_ts, word=word_text, probability=word_probability, lang=lang))

        return words



import logging
from typing import BinaryIO

import numpy as np

from .audiosource import AudioSource

LOG = logging.getLogger(__name__)

class PCMInputStreamAudioSource(AudioSource):
    stream: BinaryIO
    channels: int
    sampling_rate: int
    dtype: np.dtype
    _has_more: bool

    def __init__(self, stream: BinaryIO, channels: int = 1, sampling_rate: int = 16000, dtype=np.int16):
        super().__init__()
        self.stream = stream
        self.channels = channels
        self.sampling_rate = sampling_rate
        self.dtype = dtype
        self._has_more = True

    def next_chunk(self, chunk_length=1) -> np.ndarray:
        # Calculate the number of bytes needed per chunk
        bytes_per_sample = np.dtype(self.dtype).itemsize
        bytes_needed = chunk_length * self.sampling_rate * self.channels * bytes_per_sample

        # Buffer to store the read bytes
        input_data = bytearray()

        while len(input_data) < bytes_needed:
            chunk = self.stream.read(bytes_needed - len(input_data))
            if not chunk:
                # End of stream
                self._has_more = False
                break
            input_data.extend(chunk)

        # Convert the byte data to a NumPy array
        samples = np.frombuffer(input_data, dtype=self.dtype)
        return samples[:bytes_needed // bytes_per_sample]  # Ensure the array is the correct length

    def open(self):
        # caller is responsible for the lifecycle of the stream
        pass

    def close(self):
        # caller is responsible for the lifecycle of the stream
        pass

    def has_more(self) -> bool:
        return self._has_more



import logging
import math

import numpy as np
from pydub import AudioSegment
from scipy.signal import resample

# Configure logger
LOG = logging.getLogger(__name__)

def format_audio(audio: np.ndarray, from_sampling_rate: int) -> np.ndarray:
    to_sampling_rate = 16000

    if audio.dtype != np.float32:
        if audio.dtype == np.int8:
            audio = audio.astype(np.float32) / 128.0
        elif audio.dtype == np.int16:
            audio = audio.astype(np.float32) / 32768.0
        elif audio.dtype == np.int32:
            audio = audio.astype(np.float32) / 2147483648.0
        else:
            audio = audio.astype(np.float32)

    # If the audio is stereo, mix it down to mono
    if audio.ndim > 1:
        LOG.info(f"Mixing {audio.ndim} channels down to mono.")
        audio = np.mean(audio, axis=1)

    # Resample if the audio sample rate is not 16 kHz
    if from_sampling_rate != to_sampling_rate:
        LOG.info(f"Resampling from {from_sampling_rate} Hz to {to_sampling_rate} Hz.")
        num_samples = int(len(audio) * to_sampling_rate / from_sampling_rate)
        if num_samples == 0:
            return np.array([], dtype=np.int16)
        audio = resample(audio, num_samples)

    return audio.astype(np.float32)


def wav_to_int16(data):
    if data.dtype == np.int16:
        return data
    if data.dtype == np.float16:
        min_val = np.min(data)
        max_val = np.max(data)
        n = max(math.fabs(min_val),math.fabs(max_val))
        data = data / n
        return (data * np.iinfo(np.int16).max).astype(np.int16)
    if data.dtype == np.float32:
        min_val = np.min(data)
        max_val = np.max(data)
        n = max(math.fabs(min_val),math.fabs(max_val))
        data = data / n
        return (data * np.iinfo(np.int16).max).astype(np.int16)
    if data.dtype == np.float64:
        min_val = np.min(data)
        max_val = np.max(data)
        n = max(math.fabs(min_val),math.fabs(max_val))
        data = data / n
        return (data * np.iinfo(np.int16).max).astype(np.int16)
    if data.dtype == np.int8:
        return (data * ((1.0 * np.iinfo(np.int16).max) / np.iinfo(np.int8).max)).astype(np.int16)
    if data.dtype == np.int32:
        return (data * ((1.0 * np.iinfo(np.int16).max) / np.iinfo(np.int32).max)).astype(np.int16)
    raise ValueError(f"unexpected: {data.dtype}")

def samples_to_seconds(index:int):
    return index / 16000.0

def convert_mp3_to_wav(input_mp3, output_wav):
    # Load the mp3 file
    audio = AudioSegment.from_mp3(input_mp3)
    # Export the audio as wav
    audio.export(output_wav, format="wav")



"""
__init__.py
"""
import warnings
import logging
from matplotlib._api.deprecation import MatplotlibDeprecationWarning

# see https://github.com/asteroid-team/torch-audiomentations/issues/172
warnings.filterwarnings("ignore", category=UserWarning, module=r"torch_audiomentations\.utils\.io")

# see https://github.com/asteroid-team/torch-audiomentations/issues/172
warnings.filterwarnings("ignore", category=UserWarning, module=r".*")

warnings.filterwarnings("ignore", category=DeprecationWarning, module=r"silero_vad\.model")

warnings.filterwarnings("ignore", category=FutureWarning)

warnings.filterwarnings("ignore", category=MatplotlibDeprecationWarning)

# disable INFO:speechbrain.utils.quirks:Applied quirks (see `speechbrain.utils.quirks`): [disable_jit_profiling, allow_tf32]
#         INFO:speechbrain.utils.quirks:Excluded quirks specified by the `SB_DISABLE_QUIRKS` environment (comma-separated list): []
logging.getLogger("speechbrain.utils.quirks").setLevel(logging.WARNING)

__version__ = "1.0.2"



from abc import abstractmethod

import numpy as np

class AudioSource:
    def __init__(self):
        pass

    @abstractmethod
    def next_chunk(self, chunk_length=1) -> np.ndarray:
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def open(self):
        pass

    @abstractmethod
    def has_more(self) -> bool:
        pass



from dataclasses import dataclass
from typing import TextIO, Union, List

from colorama import Fore, Style

from .writer import TranscriptWriter, TranscriptWriterConfig
from ..formatting import format_milliseconds
from ..words import VerbatimUtterance, VerbatimWord


@dataclass
class ColorScheme:
    color_timestamp:str = Fore.CYAN
    color_speaker:str = Fore.BLUE
    color_language:str = Fore.YELLOW
    color_text:str = Fore.GREEN
    color_reset:str = Style.RESET_ALL

COLORSCHEME_ACKNOWLEDGED = ColorScheme(
    color_timestamp = Fore.LIGHTCYAN_EX,
    color_speaker = Fore.LIGHTBLUE_EX,
    color_language = Fore.LIGHTYELLOW_EX,
    color_text = Fore.LIGHTGREEN_EX,
    color_reset = Style.RESET_ALL
)

COLORSCHEME_UNACKNOWLEDGED = ColorScheme(
    color_timestamp=Fore.CYAN,
    color_speaker=Fore.BLUE,
    color_language=Fore.YELLOW,
    color_text=Fore.GREEN,
    color_reset=Style.RESET_ALL
)

COLORSCHEME_UNCONFIRMED = ColorScheme(
    color_timestamp=Fore.CYAN,
    color_speaker=Fore.BLUE,
    color_language=Fore.YELLOW,
    color_text=Fore.LIGHTBLACK_EX,
    color_reset=Style.RESET_ALL
)

COLORSCHEME_NONE = ColorScheme(
    color_timestamp = '',
    color_speaker = '',
    color_language = '',
    color_text = '',
    color_reset = ''
)

class TranscriptFormatter:
    def __init__(self):
        self.current_language = None
        self.current_speaker = None

    def format_utterance(self, utterance:VerbatimUtterance, out:TextIO, colours:ColorScheme):
        line:str = ""
        line += colours.color_timestamp
        line += f"[{format_milliseconds(utterance.start_ts * 1000 / 16000)}-{format_milliseconds(utterance.end_ts * 1000 / 16000)}]"
        if utterance.speaker != self.current_speaker:
            self.current_speaker = utterance.speaker
            line += colours.color_speaker
            line += f"[{utterance.speaker}]"
        line += colours.color_text
        for w in utterance.words:
            if w.lang != self.current_language:
                line += colours.color_language
                line += f"[{w.lang}]"
                line += colours.color_text
                self.current_language = w.lang
            line += w.word
        line += colours.color_reset
        line += '\n'
        out.write(line)

class TextIOTranscriptWriter(TranscriptWriter):
    def __init__(self, *, config: TranscriptWriterConfig, out:Union[None,TextIO],
                 acknowledged_colours=COLORSCHEME_ACKNOWLEDGED,
                 unacknowledged_colours=COLORSCHEME_UNACKNOWLEDGED,
                 unconfirmed_colors=COLORSCHEME_UNCONFIRMED,
                 print_unacknowledged:bool = False):
        super().__init__(config)
        self.formatter:TranscriptFormatter = TranscriptFormatter()
        self.out:TextIO = out
        self.acknowledged_colours = acknowledged_colours
        self.unacknowledged_colours = unacknowledged_colours
        self.unconfirmed_colors = unconfirmed_colors
        self.print_unacknowledged = print_unacknowledged

    def open(self, path_no_ext:str):
        pass

    def close(self):
        pass

    def write(self,
              utterance:VerbatimUtterance,
              unacknowledged_utterance:List[VerbatimUtterance] = None,
              unconfirmed_words:List[VerbatimWord] = None):
        self.formatter.format_utterance(utterance=utterance, out=self.out, colours=self.acknowledged_colours)
        if self.print_unacknowledged:
            if unacknowledged_utterance:
                for unack in unacknowledged_utterance:
                    self.formatter.format_utterance(utterance=unack, out=self.out, colours=self.unconfirmed_colors)
            if unconfirmed_words and len(unconfirmed_words) > 0:
                self.formatter.format_utterance(
                    utterance=VerbatimUtterance.from_words(unconfirmed_words),
                    out=self.out, colours=self.unconfirmed_colors)
        self.out.flush()

class TextTranscriptWriter(TextIOTranscriptWriter):
    def __init__(self, config: TranscriptWriterConfig):
        super().__init__(config=config, out=None,
                         acknowledged_colours=COLORSCHEME_NONE,
                         unacknowledged_colours=COLORSCHEME_NONE,
                         unconfirmed_colors=COLORSCHEME_NONE)

    def open(self, path_no_ext:str):
        # pylint: disable=consider-using-with
        self.out = open(f"{path_no_ext}.txt", "w", encoding="utf-8")

    def close(self):
        self.out.close()



from typing import Optional
import logging

import numpy as np
import av
# pylint: disable=no-name-in-module
from av.audio.resampler import AudioResampler
from av.audio.frame import AudioFrame


from .audiosource import AudioSource

LOG = logging.getLogger(__name__)

class PyAVAudioSource(AudioSource):
    """
    A chunk-based audio reader that uses PyAV to decode audio frames.

    - Reads from a container (any format that FFmpeg/PyAV supports).
    - Decodes audio frames into an internal buffer.
    - Returns them in "chunk_length" second increments via next_chunk().
    - By default, it streams as 16-bit int PCM at some target rate (e.g. 16 kHz).
    """

    def __init__(
        self,
        file_path: str,
        target_sample_rate: int = 16000,
        start_time: float = 0.0,
        end_time: Optional[float] = None,
    ):
        """
        :param file_path: Path/URL to audio file
        :param target_sample_rate: Desired sample rate for output (e.g. 16k)
        :param start_time: Seek to this time (seconds) before reading
        :param end_time: Stop reading after this time (seconds) from start
        """
        super().__init__()
        self.file_path = file_path
        self.target_sample_rate = target_sample_rate
        self.start_time = start_time
        self.end_time = end_time

        # Internals
        self._container: av.container.InputContainer = None
        # pylint: disable=c-extension-no-member
        self._stream: av.audio.stream.AudioStream = None
        self._frame_iter = None

        # Buffer for leftover samples (when frames don't line up exactly with chunk size)
        self._sample_buffer = np.array([], dtype=np.float32)

        self._closed = False
        self._done_decoding = False

    def open(self):
        """Open the container, find the audio stream, and seek if needed."""
        LOG.info(f"Opening file with PyAV: {self.file_path}")
        self._container = av.open(self.file_path)

        # Find the first audio stream (or choose a specific one if needed)
        audio_streams = [s for s in self._container.streams if s.type == "audio"]
        if not audio_streams:
            raise ValueError("No audio streams found in file.")

        self._stream = audio_streams[0]
        self._stream.thread_type = "AUTO"  # allow FFmpeg to use threading if beneficial

        # If you want to force a certain sample format, channel layout, etc.,
        # you can set those on self._stream or handle them in code below.
        # E.g., self._stream.codec_context.sample_rate = self.target_sample_rate

        # If we want to seek to `start_time`, do so in seconds:
        if self.start_time > 0:
            # av.seek() or container.seek() uses timestamps in AV_TIME_BASE units
            # but PyAV usually accepts "seconds" if we specify 'any' for backward flag
            self._container.seek(int(self.start_time / self._stream.time_base), any_frame=False, stream=self._stream)
            LOG.info(f"Seeking to {self.start_time} seconds.")

        # We create a generator that decodes frames from the audio stream
        # This is the raw frames from the container
        self._frame_iter = self._container.decode(self._stream)
        self._done_decoding = False
        self._closed = False

    def next_chunk(self, chunk_length=1) -> np.ndarray:
        """
        Return `chunk_length` seconds of audio as a NumPy array, shape (N,) or (N,1).
        If not enough data is available, decode more frames from PyAV.
        If at EOF, returns an empty array.
        """
        if self._closed:
            LOG.warning("next_chunk() called after close(). Returning empty array.")
            return np.array([], dtype=np.float32)

        if self._done_decoding and len(self._sample_buffer) == 0:
            # No more data left
            return np.array([], dtype=np.float32)

        # How many samples do we need?
        needed_samples = int(chunk_length * self.target_sample_rate)

        resampler = AudioResampler(
            format='flt',     # float32
            layout='mono',    # 1 channel
            rate=16000        # target samplerate
        )

        # Keep reading frames from PyAV until we have enough
        while len(self._sample_buffer) < needed_samples and not self._done_decoding:
            try:
                frame = next(self._frame_iter)
            except StopIteration:
                # No more frames from the container
                LOG.info("Reached end of audio stream.")
                self._done_decoding = True
                break

            # Optionally, we can check the frame's pts (presentation timestamp)
            # to see if we've passed end_time. If so, stop decoding.
            if self.end_time is not None and frame.pts is not None:
                # Convert pts to seconds
                # In PyAV, each stream has time_base -> frame_time = frame.pts * stream.time_base
                current_time_sec = float(frame.pts * self._stream.time_base)
                if current_time_sec > self.end_time:
                    LOG.info(f"Reached end_time={self.end_time:.2f}s (current={current_time_sec:.2f}s). Stopping.")
                    self._done_decoding = True
                    break

            # Resample to your desired format
            new_frames:list[AudioFrame] = resampler.resample(frame)
            for new_frame in new_frames:
                new_frame = new_frame.to_ndarray().astype(np.float32, copy=False).squeeze()
                self._sample_buffer = np.concatenate([self._sample_buffer, new_frame])

        # By now, we either have enough samples or we hit EOF
        audio_array = self._sample_buffer[:needed_samples]
        self._sample_buffer = self._sample_buffer[needed_samples:]

        # For consistency, let's return a float32 numpy array shape (N,)
        audio_array = audio_array.astype(np.float32, copy=False)

        return audio_array

    def has_more(self) -> bool:
        """
        Return True if there's more data to read from the stream,
        or if there's still leftover samples in the buffer.
        """
        if self._closed:
            return False
        if not self._done_decoding:
            return True
        # If done decoding, but still have samples in buffer, it's True
        return len(self._sample_buffer) > 0

    def close(self):
        """Close the container and release resources."""
        if self._container and not self._closed:
            LOG.info("Closing PyAV container.")
            self._container.close()
        self._closed = True
        self._done_decoding = True
        self._sample_buffer = np.array([], dtype=np.float32)



import dataclasses
from abc import abstractmethod
from typing import List
from enum import Enum

from ..words import VerbatimUtterance, VerbatimWord

# pylint: disable=invalid-name
class TimestampStyle(Enum):
    none = 1
    start = 2
    range = 3
    minute = 4

class SpeakerStyle(Enum):
    none = 1
    change = 2
    always = 3

class ProbabilityStyle(Enum):
    none = 1
    line = 2
    line_75 = 3
    line_50 = 4
    line_25 = 5
    word = 6
    word_75 = 7
    word_50 = 8
    word_25 = 9

class LanguageStyle(Enum):
    none = 1
    change = 2
    always = 3

@dataclasses.dataclass
class TranscriptWriterConfig:
    timestamp_style:TimestampStyle = TimestampStyle.none
    speaker_style:SpeakerStyle = SpeakerStyle.none
    probability_style:ProbabilityStyle = ProbabilityStyle.none
    language_style:LanguageStyle = LanguageStyle.none
    verbose:bool = False

class TranscriptWriter:
    def __init__(self, config:TranscriptWriterConfig):
        self.config = config

    @abstractmethod
    def open(self, path_no_ext:str):
        pass

    @abstractmethod
    def close(self):
        pass

    @abstractmethod
    def write(self,
              utterance:VerbatimUtterance,
              unacknowledged_utterance:List[VerbatimUtterance] = None,
              unconfirmed_words:List[VerbatimWord] = None):
        pass



import logging
import os
import wave
from typing import Dict, Union

import numpy as np
from pyannote.core.annotation import Annotation

from .audiosource import AudioSource
from ..audio import format_audio
from ..audio import convert_mp3_to_wav
from ...voices.diarization import Diarization
from ...voices.isolation import VoiceSeparator

LOG = logging.getLogger(__name__)


class FileAudioSource(AudioSource):
    diarization:Annotation
    speaker_audio:Dict[str, np.array]
    stream:wave.Wave_read = None

    def __init__(self, file: str, start_sample:int = 0, end_sample:Union[None, int] = None):
        super().__init__()
        self.file_path = file
        if self.file_path.endswith(".mp3"):
            # Convert mp3 to wav
            wav_file_path = self.file_path.replace(".mp3", ".wav")
            convert_mp3_to_wav(self.file_path, wav_file_path)
            self.file_path = wav_file_path
        self.end_sample = end_sample
        self.start_sample = start_sample

    def compute_diarization(self, device:str, rttm_file:str = None, nb_speakers:int = None) -> Annotation:
        diarization = None
        try:
            diarization = Diarization(device=device, huggingface_token=os.getenv("HUGGINGFACE_TOKEN"))
            if nb_speakers == 0:
                nb_speakers = None
            annotation = diarization.compute_diarization(file_path=self.file_path, out_rttm_file=rttm_file, nb_speakers=nb_speakers)
            return annotation
        finally:
            if diarization:
                del diarization

    def isolate_voices(self, out_path_prefix:str=None):
        LOG.info("Initializing Voice Isolation Model.")
        voice_separator:Union[None, VoiceSeparator] = None
        try:
            voice_separator = VoiceSeparator(log_level=LOG.level)
            if not out_path_prefix:
                basename, _ = os.path.splitext(os.path.basename(self.file_path))
                voice_prefix = f"{basename}-voice"
                noise_prefix = f"{basename}-noise"
            else:
                basename, ext = os.path.splitext(out_path_prefix)
                if ext:
                    voice_prefix = basename
                    noise_prefix = f"{basename}-noise"
                else:
                    voice_prefix = f"{basename}-voice"
                    noise_prefix = f"{basename}-noise"

            self.file_path, _ = voice_separator.isolate_voice_in_file(
                file=self.file_path,
                out_voice=voice_prefix,
                out_noise=noise_prefix,
                )
        finally:
            if voice_separator:
                del voice_separator

    def setpos(self, new_sample_pos:int):
        file_samplerate = self.stream.getframerate()
        if file_samplerate != 16000:
            file_sample_pos = new_sample_pos * file_samplerate // 16000
        else:
            file_sample_pos = new_sample_pos
        self.stream.setpos(int(file_sample_pos))

    def next_chunk(self, chunk_length=1) -> np.ndarray:
        LOG.info(f"Reading {chunk_length} seconds of audio from file.")
        frames = self.stream.readframes(int(self.stream.getframerate() * chunk_length))
        sample_width = self.stream.getsampwidth()
        n_channels = self.stream.getnchannels()
        dtype = np.int16 if sample_width == 2 else np.int32 if sample_width == 4 else np.uint8
        audio_array = np.frombuffer(frames, dtype=dtype)
        audio_array = audio_array.reshape(-1, n_channels)

        if len(audio_array) == 0:
            return audio_array

        audio_array = format_audio(audio_array, from_sampling_rate=self.stream.getframerate())

        LOG.info("Finished reading audio chunk from file.")
        return audio_array

    def has_more(self):
        current_frame = self.stream.tell()
        if self.end_sample is not None and current_frame > self.end_sample:
            return False
        total_frames = self.stream.getnframes()
        return current_frame < total_frames

    def open(self):
        self.stream = wave.open(self.file_path, 'rb')
        if self.start_sample != 0:
            self.setpos(self.start_sample)

    def close(self):
        self.stream.close()



import os
import sys
import unittest

from verbatim.config import Config
from verbatim.verbatim import Verbatim

sys.path.append(os.path.join(os.path.dirname(__file__), '..'))

class TestStringMethods(unittest.TestCase):
    def test_verbatim(self):
        pass

if __name__ == '__main__':
    import logging
    logging.basicConfig(level=logging.DEBUG,
                        format='%(asctime)s [%(levelname)s][%(filename)s:%(lineno)d][%(funcName)s] %(message)s',
                        datefmt='%Y-%m-%dT%H:%M:%SZ')

    os.environ['CUDA_VISIBLE_DEVICES'] = '-1'  # Set CUDA_VISIBLE_DEVICES to -1 to use CPU

    print(os.getcwd())
    config:Config = Config(input_source="tests/data/init.mp3", use_cpu=True).configure_languages(["fr", "en"])
    verbatim:Verbatim = Verbatim(config=config)
    for utterance in verbatim.transcribe():
        print(utterance.text)



import argparse
import logging
import os
import sys

from .__init__ import __version__
from .config import Config
from .transcript.format.writer import (
    TranscriptWriter,
    SpeakerStyle,
    TimestampStyle,
    ProbabilityStyle,
    LanguageStyle
)

LOG = logging.getLogger(__name__)

# Get the package name dynamically
PACKAGE_NAME = os.path.basename(os.path.dirname(os.path.abspath(__file__)))

def load_env_file(env_path=".env"):
    """
    Load environment variables from a .env file.

    Parameters:
    - env_path (str): The path to the .env file. Default is '.env'.

    Returns:
    - bool: True if the file was successfully loaded, False otherwise.
    """
    if not os.path.isfile(env_path):
        print(f"File '{env_path}' does not exist.")
        return False

    try:
        with open(env_path, "r", encoding="utf-8") as file:
            for line in file:
                line = line.strip()
                # Skip empty lines and comments
                if not line or line.startswith("#"):
                    continue

                # Ensure the line contains a key-value pair
                if "=" not in line:
                    print(f"Ignored invalid line: '{line}'")
                    continue

                # Split the key and value
                key, value = line.split("=", 1)
                key = key.strip()
                value = value.strip()

                # Set the environment variable
                os.environ[key] = value

        LOG.info(f"Environment variables from '{env_path}' loaded successfully.")
        return True
    # pylint: disable=broad-exception-caught
    except Exception as e:
        print(f"Error while loading '{env_path}': {e}")
        return False

def configure_writers(config:Config, original_audio_file:str) -> TranscriptWriter:
    # pylint: disable=import-outside-toplevel
    from .transcript.format import (TextTranscriptWriter, MultiTranscriptWriter, AssTranscriptWriter,
        DocxTranscriptWriter, MarkdownTranscriptWriter, JsonTranscriptWriter, StdoutTranscriptWriter)

    multi_writer:MultiTranscriptWriter = MultiTranscriptWriter()
    if config.enable_txt:
        multi_writer.add_writer(TextTranscriptWriter(config=config.write_config))
    if config.enable_ass:
        multi_writer.add_writer(AssTranscriptWriter(config=config.write_config, original_audio_file=original_audio_file))
    if config.enable_docx:
        multi_writer.add_writer(DocxTranscriptWriter(config=config.write_config))
    if config.enable_md:
        multi_writer.add_writer(MarkdownTranscriptWriter(config=config.write_config))
    if config.enable_json:
        multi_writer.add_writer(JsonTranscriptWriter(config=config.write_config))
    if config.enable_stdout and not config.enable_stdout_nocolor:
        multi_writer.add_writer(StdoutTranscriptWriter(config=config.write_config, with_colours=True))
    if config.enable_stdout_nocolor:
        multi_writer.add_writer(StdoutTranscriptWriter(config=config.write_config, with_colours=False))
    return multi_writer

def main():
    class OptionalValueAction(argparse.Action):
        def __call__(self, parser, namespace, values, option_string=None):
            # Set the attribute to the provided value or an empty string if no value is given
            setattr(namespace, self.dest, values if values is not None else "")

    parser = argparse.ArgumentParser(prog=PACKAGE_NAME, description="Verbatim: that's what she said")

    # Specify the command line arguments
    parser.add_argument("input", nargs='?', default=None,
                        help="Path to the input audio file. Use '-' to read from stdin "
                        "(expecting 16bit 16kHz mono PCM stream) or '>' to use microphone.")
    parser.add_argument("-f", "--from",
                        help="Start time within the file in hh:mm:ss.ms or mm:ss.ms", default="00:00.000", dest="start_time")
    parser.add_argument("-t", "--to",
                        help="Stop time within the file in hh:mm:ss.ms or mm:ss.ms", default="", dest="stop_time")
    parser.add_argument("-o", "--outdir",
                        help="Path to the output directory", default=".")
    parser.add_argument("-d", "--diarization", nargs='?', action=OptionalValueAction, default=None,
                        help="Identify speakers in transcript using the diarization RTTM file at the specified path (ex. diarization.rttm)")
    parser.add_argument("-i", "--isolate", nargs='?', action=OptionalValueAction, default=None,
                    help="Extract voices from background noise. Outputs files <name>-vocals.wav and <name>-noise.wav "
                    "if a name is provided, otherwise uses default names.")
    parser.add_argument("-l", "--languages", nargs="*",
                        help="Languages for speech recognition. Provide multiple values for multiple languages.")
    parser.add_argument("-n", "--diarize", nargs='?', action=OptionalValueAction, default=None,
                        help="Number of speakers in the audio file.")
    parser.add_argument("-b", "--nb_beams", type=int,
                        help="Number of parallel when resolving transcription. " +
                        "1-3 for fast, 12-15 for high quality. Default is 9.",
                        default=None)
    parser.add_argument("-v", "--verbose", action="count", default=0,
                        help="Increase verbosity (specify multiple times for more verbosity)")
    parser.add_argument("--version", action="version", version=f"{PACKAGE_NAME} {__version__}")
    parser.add_argument("--cpu", action="store_true", help="Toggle CPU usage")
    parser.add_argument("-s", "--stream", action="store_true", help="Set mode to low latency streaming")
    parser.add_argument("-w", "--workdir", nargs='?', action=OptionalValueAction, default=None,
                        help="Set the working directory where temporary files may be written to (default is system temp directory)")
    parser.add_argument("--ass", action="store_true", help="Enable ASS subtitle file output")
    parser.add_argument("--docx", action="store_true", help="Enable Microsoft Word DOCX output")
    parser.add_argument("--txt", action="store_true", help="Enable TXT file output")
    parser.add_argument("--json", action="store_true", help="Enable json file output")
    parser.add_argument("--md", action="store_true", help="Enable Markdown (MD) output")
    parser.add_argument("--stdout", action="store_true", default=True, help="Enable stdout output (enabled by default)")
    parser.add_argument("--stdout-nocolor", action="store_true", help="Enable stdout output without colors")
    parser.add_argument("--format-timestamp", type=lambda s: TimestampStyle[s], choices=list(TimestampStyle), default=TimestampStyle.minute,
                        help="Set the timestamp format: 'none' for no timestamps, 'start' for start time, 'range' for start and end times")
    parser.add_argument("--format-speaker", type=lambda s: SpeakerStyle[s], choices=list(SpeakerStyle), default=SpeakerStyle.change,
                        help="Set the timestamp format: 'none' for no timestamps, 'start' for start time, 'range' for start and end times")
    parser.add_argument("--format-probability", type=lambda s: ProbabilityStyle[s], choices=list(ProbabilityStyle), default=ProbabilityStyle.line,
                        help="Set the timestamp format: 'none' for no timestamps, 'start' for start time, 'range' for start and end times")
    parser.add_argument("--format-language", type=lambda s: LanguageStyle[s], choices=list(LanguageStyle), default=LanguageStyle.change,
                        help="Set the timestamp format: 'none' for no timestamps, 'start' for start time, 'range' for start and end times")

    args = parser.parse_args()
    # Set logging level based on verbosity
    log_levels = [logging.WARNING, logging.INFO, logging.DEBUG]
    log_level = log_levels[min(args.verbose, len(log_levels) - 1)]
    for handler in logging.root.handlers[:]:
        logging.root.removeHandler(handler)
    logging.basicConfig(stream=sys.stderr,
                        level=log_level,
                        format='%(asctime)s [%(levelname)s][%(filename)s:%(lineno)d][%(funcName)s] %(message)s',
                        datefmt='%Y-%m-%dT%H:%M:%SZ')

    load_env_file()

    # Check if the version option is specified
    if hasattr(args, 'version') and args.version:
        return  # Exit if the version option is specified

    config:Config = Config(
        use_cpu=args.cpu,
        input_source=args.input,
        outdir=args.outdir,
        workdir=args.workdir,
        stream=args.stream,
        isolate=args.isolate,
        diarize=args.diarize,
        start_time=args.start_time,
        stop_time=args.stop_time)

    config.lang = args.languages if args.languages else ["en"]

    # Set output formats
    config.write_config.timestamp_style = args.format_timestamp
    config.write_config.probability_style = args.format_probability
    config.write_config.speaker_style = args.format_speaker
    config.write_config.language_style = args.format_language
    config.write_config.verbose = log_level <= logging.INFO

    output_formats = []
    if args.ass:
        output_formats.append("ass")
    if args.docx:
        output_formats.append("docx")
    if args.txt:
        output_formats.append("txt")
    if args.md:
        output_formats.append("md")
    if args.json:
        output_formats.append("json")
    if args.stdout:
        output_formats.append("stdout")
    if args.stdout_nocolor:
        output_formats.append("stdout-nocolor")

    config.configure_output_formats(output_formats=output_formats)

    # pylint: disable=import-outside-toplevel
    from .verbatim import Verbatim
    transcriber = Verbatim(config)
    writer:TranscriptWriter = configure_writers(config, original_audio_file=config.input_source)
    writer.open(path_no_ext=config.output_prefix_no_ext)
    for utterance, unacknowledged, unconfirmed in transcriber.transcribe():
        writer.write(utterance=utterance, unacknowledged_utterance=unacknowledged, unconfirmed_words=unconfirmed)
    writer.close()

if __name__ == "__main__":
    sys.argv = ['run.py', 'samples/voices.wav', '-w', 'out', '--language', 'en', 'fr', '--txt', '--md', '--json', '--docx', '--ass']
    main()



import logging
import os
import re
import errno

from dataclasses import dataclass, field
from typing import Dict, List, Union

from pyannote.core.annotation import Annotation

from .audio.sources.audiosource import AudioSource
from .transcript.format.writer import TranscriptWriterConfig

LOG = logging.getLogger(__name__)

def timestr_to_sample(timestr: str, sample_rate: int = 16000) -> int:
    """
    Converts a time string in the format hh:mm:ss.ms, mm:ss.ms, or ss.ms
    (milliseconds optional) to the corresponding sample index.

    Args:
        timestr (str): Time string in the format hh:mm:ss.ms, mm:ss.ms, or ss.ms.
        sample_rate (int): Sampling rate in Hz (default is 16000).

    Returns:
        int: The corresponding sample index.
    """
    # Define regex patterns for specific formats
    hh_mm_ss_ms_pattern = re.compile(
        r"^(?P<hours>\d+):(?P<minutes>\d+):(?P<seconds>\d+)(?:\.(?P<milliseconds>\d+))?$"
    )
    mm_ss_ms_pattern = re.compile(
        r"^(?P<minutes>\d+):(?P<seconds>\d+)(?:\.(?P<milliseconds>\d+))?$"
    )
    ss_ms_pattern = re.compile(
        r"^(?P<seconds>\d+)(?:\.(?P<milliseconds>\d+))?$"
    )

    # Match the input against patterns
    if match := hh_mm_ss_ms_pattern.match(timestr.strip()):
        hours = int(match.group("hours"))
        minutes = int(match.group("minutes"))
        seconds = int(match.group("seconds"))
        milliseconds = int(match.group("milliseconds")) if match.group("milliseconds") else 0
    elif match := mm_ss_ms_pattern.match(timestr.strip()):
        hours = 0
        minutes = int(match.group("minutes"))
        seconds = int(match.group("seconds"))
        milliseconds = int(match.group("milliseconds")) if match.group("milliseconds") else 0
    elif match := ss_ms_pattern.match(timestr.strip()):
        hours = 0
        minutes = 0
        seconds = int(match.group("seconds"))
        milliseconds = int(match.group("milliseconds")) if match.group("milliseconds") else 0
    else:
        raise ValueError(f"Invalid time string format: {timestr}")

    # Calculate total time in seconds
    total_seconds = hours * 3600 + minutes * 60 + seconds + milliseconds / 1000.0

    # Convert to sample index
    return int(total_seconds * sample_rate)


@dataclass
class Config:
    sampling_rate: int = 16000
    transcribe_latency: int = 16000
    frames_per_buffer: int = 1000
    window_duration: int = 30  # seconds
    whisper_prompts: Dict[str, str] = None
    lang: List[str] = field(default_factory=list)
    start_time:int = 0
    stop_time:Union[None,int] = None
    source_stream: AudioSource = None
    input_source:str = None
    isolate:Union[None,bool] = None
    diarization: Annotation = None
    diarization_file: str = None
    device: str = "cuda"
    stream: bool = False
    diarize:Union[int,None] = None
    whisper_beam_size: int = 9
    whisper_best_of: int = 9
    whisper_patience: float = 2.0
    whisper_temperatures: List[float] = None
    debug: bool = False
    working_dir: str = "."
    output_dir: str = "."
    enable_ass: bool = False
    enable_docx: bool = False
    enable_txt: bool = False
    enable_md: bool = False
    enable_json: bool = False
    enable_stdout: bool = True
    enable_stdout_nocolor: bool = False
    write_config:TranscriptWriterConfig = field(default=TranscriptWriterConfig)
    output_prefix_no_ext:str = "out"
    working_prefix_no_ext:str = "out"

    def __init__(
            self, *,
            input_source:Union[None,str],
            outdir:Union[None,str] = ".", workdir:Union[None,str] = None,
            use_cpu:Union[None, bool] = None,
            stream:Union[None,bool] = False, isolate:Union[None,bool]=None, diarize:Union[None,int] = None,
            start_time:Union[None,str] = "0", stop_time:Union[None,str] = None
            ):

        self.chunk_table = [
            (0.75, 0.20),
            (0.50, 0.15),
            (0.25, 0.10),
            (0.10, 0.05),
            (0.00, 0.025),
        ]
        self.lang = ["en"]
        self.whisper_prompts = {
            "en": "This is a sentence.",
            "zh": "这是一个句子。",
            "de": "Dies ist ein Satz.",
            "es": "Esta es una oración.",
            "ru": "Это предложение.",
            "ko": "이것은 문장입니다.",
            "fr": "Ceci est une phrase.",
            "ja": "これは文です。",
            "pt": "Esta é uma frase.",
            "tr": "Bu bir cümledir.",
            "pl": "To jest zdanie.",
            "ca": "Això és una oració.",
            "nl": "Dit is een zin.",
            "ar": "هذه جملة.",
            "sv": "Det här är en mening.",
            "it": "Questa è una frase.",
            "id": "Ini adalah kalimat.",
            "hi": "यह एक वाक्य है।",
            "fi": "Tämä on lause.",
            "vi": "Đây là một câu.",
            "he": "זה משפט.",
            "uk": "Це речення.",
            "el": "Αυτή είναι μια πρόταση.",
            "ms": "Ini ialah ayat.",
            "cs": "Toto je věta.",
            "ro": "Aceasta este o propoziție.",
            "da": "Dette er en sætning.",
            "hu": "Ez egy mondat.",
            "ta": "இது ஒரு வாக்கியம்.",
            "no": "Dette er en setning.",
            "th": "นี่คือประโยค.",
            "ur": "یہ ایک جملہ ہے۔",
            "hr": "Ovo je rečenica.",
            "bg": "Това е изречение.",
            "lt": "Tai yra sakinys.",
            "la": "Haec sententia est.",
            "mi": "He rerenga kōrero tēnei.",
            "ml": "ഇത് ഒരു വാക്യം ആണ്.",
            "cy": "Dyma frawddeg.",
            "sk": "Toto je veta.",
            "te": "ఇది ఒక వాక్యం.",
            "fa": "این یک جمله است.",
            "lv": "Šis ir teikums.",
            "bn": "এটি একটি বাক্য।",
            "sr": "Ово је реченица.",
            "az": "Bu bir cümlədir.",
            "sl": "To je stavek.",
            "kn": "ಇದು ಒಂದು ವಾಕ್ಯ.",
            "et": "See on lause.",
            "mk": "Ова е реченица.",
            "br": "Hemañ zo ur frazenn.",
            "eu": "Hau esaldi bat da.",
            "is": "Þetta er setning.",
            "hy": "Սա նախադասություն է։",
            "ne": "यो एउटा वाक्य हो।",
            "mn": "Энэ бол өгүүлбэр юм.",
            "bs": "Ovo je rečenica.",
            "kk": "Бұл сөйлем.",
            "sq": "Kjo është një fjali.",
            "sw": "Hii ni sentensi.",
            "gl": "Esta é unha oración.",
            "mr": "हे एक वाक्य आहे.",
            "pa": "ਇਹ ਇੱਕ ਵਾਕ ਹੈ।",
            "si": "මේ වාක්‍යයක් වේ.",
            "km": "នេះជាប្រយោគមួយ។",
            "sn": "Uyu mutsara.",
            "yo": "Ọ̀rọ̀ yìí ni.",
            "so": "Tani waa jumlad.",
            "af": "Dit is 'n sin.",
            "oc": "Aquò es una frasa.",
            "ka": "ეს წინადადებაა.",
            "be": "Гэта сказ.",
            "tg": "Ин ҷумла аст.",
            "sd": "هيءَ هڪ جملو آهي.",
            "gu": "આ એક વાક્ય છે.",
            "am": "ይህ አንድ ዓረፍተ ነገር ነው።",
            "yi": "דאָס איז אַ זאַץ.",
            "lo": "ນີ້ແມ່ນປະໂຫຍກ.",
            "uz": "Bu bir gap.",
            "fo": "Hetta er ein setning.",
            "ht": "Sa a se yon fraz.",
            "ps": "دا یوه جمله ده.",
            "tk": "Bu bir sözlemdir.",
            "nn": "Dette er ei setning.",
            "mt": "Din hija sentenza.",
            "sa": "एषा एकं वाक्यम् अस्ति।",
            "lb": "Dëst ass e Saz.",
            "my": "ဒါက စာကြောင်းတစ်ခုပါ။",
            "bo": "འདི་ནི་ཚིག་གྲུབ་ཞིག་རེད།",
            "tl": "Ito ay isang pangungusap.",
            "mg": "Io dia fehezanteny.",
            "as": "এইটো এটা বাক্য।",
            "tt": "Бу җөмлә.",
            "haw": "He hopunaʻōlelo kēia.",
            "ln": "Oyo ezali lokasa.",
            "ha": "Wannan jumla ce.",
            "ba": "Был һөйләм.",
            "jw": "Iki ukara.",
            "su": "Ieu mangrupikeun kalimah."
        }
        self.whisper_temperatures = [0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0]
        self.working_dir = os.getenv("TMPDIR", os.getenv("TEMP", os.getenv("TMP", ".")))

        # pylint: disable=import-outside-toplevel
        import torch
        if use_cpu or not torch.cuda.is_available():
            os.environ['CUDA_VISIBLE_DEVICES'] = '-1'  # Set CUDA_VISIBLE_DEVICES to -1 to use CPU
            LOG.info("Using CPU")
            self.device = "cpu"
        else:
            LOG.info("Using GPU")
            self.device = "cuda"

        if stream:
            self.configure_for_low_latency_streaming()

        self.isolate = isolate
        self.diarize = diarize
        if self.diarize == '':
            self.diarize = 0
        elif self.diarize is not None:
            self.diarize = int(self.diarize)

        self.configure_output_directory(outdir=outdir, workdir=workdir)
        if input_source is not None:
            self.configure_audio_source(input_source=input_source, start_time=start_time, stop_time=stop_time)

    def configure_output_directory(self, outdir:str, workdir:str):
        if not os.path.isdir(outdir):
            os.makedirs(outdir)
        self.output_dir = outdir
        LOG.info(f"Output directory set to {self.output_dir}")

        # Set the working directory
        if workdir is None:
            self.working_dir = os.getenv("TMPDIR", os.getenv("TEMP", os.getenv("TMP", ".")))
        elif workdir == "":
            self.working_dir = self.output_dir
        else:
            if not os.path.isdir(workdir):
                os.makedirs(workdir)
            self.working_dir = workdir
        LOG.info(f"Working directory set to {self.working_dir}")


    def configure_audio_source(self, input_source:str, start_time:str, stop_time:str):
        self.input_source = input_source
        if os.path.exists(input_source) is False:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), input_source)

        input_name_no_ext = os.path.splitext(os.path.split(self.input_source)[-1])[0]
        self.output_prefix_no_ext = os.path.join(self.output_dir, input_name_no_ext)
        self.working_prefix_no_ext = os.path.join(self.working_dir, input_name_no_ext)

        self.diarization_file = self.diarization
        if self.diarization_file == "" or (self.diarize is not None and self.diarization_file is None):
            self.diarization_file = self.output_prefix_no_ext + ".rttm"

        self.start_time = timestr_to_sample(start_time) if start_time else None
        self.stop_time = timestr_to_sample(stop_time) if stop_time else None

        # pylint: disable=import-outside-toplevel
        from .audio.sources.micaudiosource import MicAudioSourcePyAudio as MicAudioSource
        from .audio.sources.fileaudiosource import FileAudioSource
        from .audio.sources.ffmpegfileaudiosource import PyAVAudioSource
        from .audio.sources.pcmaudiosource import PCMInputStreamAudioSource
        import sys
        import numpy as np

        if self.input_source == "-":
            self.source_stream = PCMInputStreamAudioSource(stream=sys.stdin, channels=1, sampling_rate=16000, dtype=np.int16)
            return

        elif self.input_source is None or self.input_source == ">":
            self.source_stream = MicAudioSource()
            return
        else:
            if os.path.splitext(self.input_source)[-1] == ".wav":
                file_audio_source = FileAudioSource(self.input_source, start_sample=self.start_time, end_sample=self.stop_time)
                if not self.stream:
                    if self.isolate is not None:
                        file_audio_source.isolate_voices(out_path_prefix=self.isolate or None)
                    if not self.diarize is None:
                        self.diarization = file_audio_source.compute_diarization(
                            rttm_file=self.diarization_file, device=self.device, nb_speakers=self.diarize)

                if self.diarization_file:
                    from .voices.diarization import Diarization
                    self.diarization = Diarization.load_diarization(rttm_file=self.diarization_file)
            else:
                if not self.stream and (self.isolate is not None or not self.diarize is None):
                    file_audio_source = PyAVAudioSource(file_path=self.input_source)
                    from .audio.sources.wavsink import WavSink
                    self.input_source = self.working_prefix_no_ext + '.wav'
                    WavSink.dump_to_wav(audio_source=file_audio_source, output_path=self.input_source)
                    self.configure_audio_source(input_source=self.input_source, start_time=start_time, stop_time=stop_time)
                    return
                else:
                    file_audio_source = PyAVAudioSource(
                        file_path=self.input_source, start_time=self.start_time/16000, end_time=self.stop_time/16000 if self.stop_time else None)
            self.source_stream = file_audio_source
            return

    def configure_for_low_latency_streaming(self):
        self.stream = True
        if self.stream:
            self.chunk_table = [
                (0, 0.025),
            ]
            self.whisper_best_of = 3
            self.whisper_beam_size = 3
            self.whisper_patience = 3.0
            self.whisper_temperatures = [0, 0.6]

    def configure_output_formats(self, output_formats:List[str]) -> "Config":
        self.enable_ass = "ass" in output_formats
        self.enable_docx = "docx" in output_formats
        self.enable_txt = "txt" in output_formats
        self.enable_md = "md" in output_formats
        self.enable_json = "json" in output_formats
        self.enable_stdout = "json" in output_formats
        self.enable_stdout = "stdout" in output_formats
        self.enable_stdout_nocolor = "stdout-nocolor" in output_formats
        return self

    def configure_languages(self, lang=List[str]) -> "Config":
        self.lang = lang
        return self



from typing import TextIO, Union, List

from .writer import TranscriptWriter, TranscriptWriterConfig
from ..words import VerbatimUtterance, VerbatimWord


class TranscriptFormatter:
    def __init__(self):
        self.current_language = None
        self.current_speaker = None
        self.first_utterance = True

    def open(self, out:TextIO):
        out.write("[\n")

    def close(self, out:TextIO):
        out.write("\n]")

    def format_utterance(self, utterance:VerbatimUtterance, out:TextIO):
        if self.first_utterance:
            self.first_utterance = False
        else:
            out.write(',\n')
        out.write('  {\n')
        out.write(f'    "start_sample": {utterance.start_ts}, "start_second": {utterance.start_ts / 16000:.2f},\n')
        out.write(f'    "end_sample": {utterance.end_ts}, "end_second": {utterance.end_ts / 16000:.2f},\n')
        out.write(f'    "speaker": "{utterance.speaker}",\n')
        out.write(f'    "text": "{utterance.text}",\n')
        out.write( '    "words": [')
        first_word = True
        for w in utterance.words:
            if first_word:
                first_word = False
                out.write('\n')
            else:
                out.write(',\n')
            out.write(f'      {{ "text": "{w.word}", "lang": "{w.lang}", "prob": {w.probability:.4f} }}')
        out.write('\n    ]\n')
        out.write('  }')

class JsonTranscriptWriter(TranscriptWriter):
    def __init__(self, config: TranscriptWriterConfig):
        super().__init__(config)
        self.formatter:TranscriptFormatter = TranscriptFormatter()
        self.out:Union[None,TextIO] = None

    def open(self, path_no_ext:str):
        # pylint: disable=consider-using-with
        self.out = open(f"{path_no_ext}.json", "w", encoding="utf-8")
        self.formatter.open(out=self.out)

    def close(self):
        self.formatter.close(out=self.out)
        self.out.close()

    def write(self,
              utterance:VerbatimUtterance,
              unacknowledged_utterance:List[VerbatimUtterance] = None,
              unconfirmed_words:List[VerbatimWord] = None):
        self.formatter.format_utterance(utterance=utterance, out=self.out)
        self.out.flush()



import logging
from typing import List, Tuple

import numpy as np
from faster_whisper import WhisperModel
from ...audio.audio import samples_to_seconds
from ...transcript.words import VerbatimWord
from .transcribe import Transcriber, WhisperConfig

LOG = logging.getLogger(__name__)

class FasterWhisperTranscriber(Transcriber):
    def __init__(self, *, model_size_or_path:str, device:str,
        whisper_beam_size:int = 3,
        whisper_best_of:int = 3,
        whisper_patience:float = 1.0,
        whisper_temperatures:List[float] = None
    ):
        if whisper_temperatures is None:
            whisper_temperatures = [0.0, 0.2, 0.4, 0.6, 0.8, 1.0]

        if device == "cpu":
            self.whisper_model = WhisperModel(model_size_or_path=model_size_or_path, device=device, compute_type="default", cpu_threads=16)
        else:
            self.whisper_model = WhisperModel(model_size_or_path=model_size_or_path, device=device, compute_type="float16")

        self.whisper_beam_size = whisper_beam_size
        self.whisper_best_of = whisper_best_of
        self.whisper_patience = whisper_patience
        self.whisper_temperatures = whisper_temperatures

    def guess_language(self, audio:np.array, lang:List[str]) -> Tuple[str, float]:
        language, language_probability, all_language_probs = self.whisper_model.detect_language(audio=audio)
        if language in lang:
            LOG.info(f"detected '{language}' with probability {language_probability}")
            return language, language_probability

        guess_lang = lang[0]
        guess_prob = 0
        for l in lang:
            for t in all_language_probs:
                if t[0] == l:
                    if t[1] > guess_prob:
                        guess_lang = t[0]
                        guess_prob = t[1]
                        LOG.info(f"detected '{l}' with probability {guess_prob}")
        return guess_lang, guess_prob


    def transcribe(
            self, *,
            audio:np.array,
            lang: str, prompt: str, prefix: str,
            window_ts:int, audio_ts:int,
            whisper_beam_size:int = 3,
            whisper_best_of:int = 3,
            whisper_patience:float = 1.0,
            whisper_temperatures:List[float] = None
            ) -> List[VerbatimWord]:
        LOG.info(f"Transcription Prefix: {prefix}")
        if whisper_temperatures is None:
            whisper_temperatures = [0.0, 0.2, 0.4, 0.6, 0.8, 1.0]

        show_progress = LOG.getEffectiveLevel() <= logging.INFO
        whisper_config:WhisperConfig = WhisperConfig()
        segment_iter, info = self.whisper_model.transcribe(audio=audio,
            language=lang,
            task=whisper_config.task,
            log_progress=show_progress,
            beam_size=whisper_beam_size,
            best_of=whisper_best_of,
            patience=whisper_patience,
            length_penalty=whisper_config.length_penalty, repetition_penalty=1.0,
            no_repeat_ngram_size=0,
            temperature=whisper_temperatures,
            compression_ratio_threshold=whisper_config.compression_ratio_threshold,
            log_prob_threshold=whisper_config.logprob_threshold,
            no_speech_threshold=whisper_config.no_speech_threshold,
            condition_on_previous_text=False,
            prompt_reset_on_temperature=0.0,
            initial_prompt=prompt, prefix=prefix,
            suppress_blank=whisper_config.suppress_blank, suppress_tokens=whisper_config.suppress_tokens,
            without_timestamps=False, max_initial_timestamp=1.0,
            word_timestamps=True,
            prepend_punctuations=whisper_config.prepend_punctuations,
            append_punctuations=whisper_config.append_punctuations,
            multilingual=False,
            vad_filter=False, vad_parameters=None,
            max_new_tokens=None, chunk_length=None,
            clip_timestamps="0.0",
            hallucination_silence_threshold=None,
            hotwords=None,
            language_detection_threshold=0.5,
            language_detection_segments=1)

        LOG.debug(f"info={info}")

        transcript_words: List[VerbatimWord] = []
        for segment in segment_iter:
            for w in segment.words:
                word = VerbatimWord.from_word(word=w, lang=lang, ts_offset=window_ts)
                if word.end_ts > audio_ts:
                    continue

                LOG.debug(f"[{word.start_ts} ({samples_to_seconds(word.start_ts)})]: {w.word}")
                transcript_words.append(word)

        return transcript_words



import logging
from abc import abstractmethod
from typing import Dict, List
import warnings

import numpy as np
import torch

from silero_vad import get_speech_timestamps, load_silero_vad

# Configure logger
LOG = logging.getLogger(__name__)

class VoiceActivityDetection:
    @abstractmethod
    def find_activity(self, audio: np.ndarray):
        pass

class SileroVoiceActivityDetection(VoiceActivityDetection):
    def __init__(self):
        with warnings.catch_warnings():
            # suppresses silero_vad\model.py:15: DeprecationWarning: path is deprecated. Use files() instead.
            # Refer to https://importlib-resources.readthedocs.io/en/latest/using.html#migrating-from-legacy for migration advice.
            warnings.filterwarnings("ignore", category=DeprecationWarning)
            self.model = load_silero_vad()

    def find_activity(self, audio: np.ndarray, min_speech_duration_ms:int=250, min_silence_duration_ms:int=100) -> List[Dict[str, int]]:
        audio_tensor = torch.from_numpy(audio).float()
        speech_timestamps = get_speech_timestamps(
            audio=audio_tensor,
            model=self.model,
            threshold= 0.25,
            sampling_rate = 16000,
            min_speech_duration_ms = min_speech_duration_ms,
            max_speech_duration_s = float('inf'),
            min_silence_duration_ms = min_silence_duration_ms,
            speech_pad_ms = 250,
            return_seconds = False,
            visualize_probs = False,
            progress_tracking_callback = None,
            neg_threshold = None,
            window_size_samples = 512)
        return speech_timestamps



from typing import List

from .writer import TranscriptWriter, TranscriptWriterConfig
from ..words import VerbatimUtterance, VerbatimWord


class MultiTranscriptWriter(TranscriptWriter):
    def __init__(self, writers: List[TranscriptWriter] = None):
        super().__init__(config=TranscriptWriterConfig())
        self.writers = []
        if writers:
            self.writers += writers

    def get_extension(self):
        return ""

    def open(self, path_no_ext:str):
        for w in self.writers:
            w.open(path_no_ext)

    def close(self):
        for w in self.writers:
            w.close()

    def write(self,
              utterance:VerbatimUtterance,
              unacknowledged_utterance:List[VerbatimUtterance] = None,
              unconfirmed_words:List[VerbatimWord] = None):
        for w in self.writers:
            w.write(utterance=utterance, unacknowledged_utterance=unacknowledged_utterance, unconfirmed_words=unconfirmed_words)

    def add_writer(self, writer:TranscriptWriter):
        self.writers.append(writer)



import logging
from abc import abstractmethod
from typing import List, Tuple

import numpy as np
from ...transcript.words import VerbatimWord

LOG = logging.getLogger(__name__)

PREPEND_PUNCTUATIONS="\"'“¿([{-"
APPEND_PUNCTUATIONS="\"'.。,;，!！?？:：”)]}、"


class WhisperConfig:
    def __init__(self):
        self.prepend_punctuations = PREPEND_PUNCTUATIONS
        self.append_punctuations = APPEND_PUNCTUATIONS
        self.task = "transcribe"
        self.length_penalty = 1.0
        self.suppress_tokens = [-1]
        self.suppress_blank = False
        self.no_speech_threshold = 0.95
        self.logprob_threshold = -1.0
        self.compression_ratio_threshold = 2.4

class Transcriber:
    @abstractmethod
    def guess_language(self, audio:np.array, lang:List[str]) -> Tuple[str, float]:
        pass

    @abstractmethod
    def transcribe(
        self, *,
        audio:np.array,
        lang: str, prompt: str, prefix: str,
        window_ts:int, audio_ts:int,
        whisper_beam_size:int = 3,
        whisper_best_of:int = 3,
        whisper_patience:float = 1.0,
        whisper_temperatures:List[float] = None
    ) -> List[VerbatimWord]:
        pass



from .writer import TranscriptWriter
from .ass import AssTranscriptWriter
from .docx import DocxTranscriptWriter
from .json import JsonTranscriptWriter
from .md import MarkdownTranscriptWriter
from .multi import MultiTranscriptWriter
from .txt import TextTranscriptWriter, TextIOTranscriptWriter
from .stdout import StdoutTranscriptWriter



import logging
from abc import abstractmethod
from typing import List
import re
from wtpsplit import SaT

# Configure logger
LOG = logging.getLogger(__name__)

class SentenceTokenizer:
    @abstractmethod
    def split(self, text:str) -> List[str]:
        pass

class FastSentenceTokenizer(SentenceTokenizer):
    def split(self, text:str) -> List[str]:
        # List of punctuation marks to split on
        split_punctuations = r".。;!！?？"
        punctuations = "\"'.。,;，!！?？:：”)]}、\"'“¿([{-"
        # Create a regex pattern to match any punctuation character
        regex_punctuation = f"([{re.escape(split_punctuations)}])"
        regex_not_punctuation = f"([^{re.escape(punctuations)}])"
        # Split the sentence, keeping the punctuation in the result
        parts = re.split(regex_punctuation, text)
        # Merge punctuation with the preceding part, preserving spaces
        result = []
        words = ""
        for part in parts:
            words += part
            if words and re.match(regex_not_punctuation, words):
                result.append(words)
                words = ""

        if words:
            if re.match(regex_not_punctuation, words):
                result.append(words)
            elif len(result) > 0:
                result[-1] += words
            else:
                result.append(words)

        return result

class SaTSentenceTokenizer(SentenceTokenizer):
    def __init__(self, device:str, model="sat-12l-sm"):
        self.sat_sm = SaT(model)
        self.sat_sm.half().to(device)

    def split(self, text:str) -> List[str]:
        return self.sat_sm.split(text)



import wave
import numpy as np

from .audiosource import AudioSource

class WavSink:
    @staticmethod
    def dump_to_wav(audio_source: AudioSource, output_path: str, sample_rate:int = 16000):
        """
        Dump the entire audio content from PyAVAudioSource to a .wav file.
        """
        # Open the PyAVAudioSource
        audio_source.open()

        # pylint: disable=no-member
        with wave.open(output_path, 'w') as wav_file:
            # Set WAV parameters
            num_channels = 1  # Mono
            sample_width = 2  # 16-bit PCM
            frame_rate = sample_rate  # Target sample rate
            wav_file.setnchannels(num_channels)
            wav_file.setsampwidth(sample_width)
            wav_file.setframerate(frame_rate)

            while audio_source.has_more():
                # Read a chunk of audio (default 1 second chunks)
                audio_chunk = audio_source.next_chunk(chunk_length=1)

                # Convert float32 audio to 16-bit PCM for WAV format
                int_samples = (audio_chunk * 32767).clip(-32768, 32767).astype(np.int16)

                # Write to WAV file
                wav_file.writeframes(int_samples.tobytes())

        # Close the audio source
        audio_source.close()



from .audiosource import AudioSource
from .fileaudiosource import FileAudioSource
from .micaudiosource import MicAudioSourcePyAudio as MicAudioSource
from .pcmaudiosource import PCMInputStreamAudioSource
